// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
// vi: set et ts=4 sw=4 sts=4:
/*****************************************************************************
 *   See the file COPYING for full copying permissions.                      *
 *                                                                           *
 *   This program is free software: you can redistribute it and/or modify    *
 *   it under the terms of the GNU General Public License as published by    *
 *   the Free Software Foundation, either version 2 of the License, or       *
 *   (at your option) any later version.                                     *
 *                                                                           *
 *   This program is distributed in the hope that it will be useful,         *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the            *
 *   GNU General Public License for more details.                            *
 *                                                                           *
 *   You should have received a copy of the GNU General Public License       *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.   *
 *****************************************************************************/
/*!
 * \file
 * \brief An assembler for the global Jacobian matrix for models using the box discretization.
 */
#ifndef DUMUX_BOX_ASSEMBLER_HH
#define DUMUX_BOX_ASSEMBLER_HH

#include <dune/common/version.hh>

#include <dumux/common/exceptions.hh>
#include <dumux/implicit/common/implicitassembler.hh>
#include <dumux/parallel/vertexhandles.hh>

namespace Dumux {

/*!
 * \ingroup BoxModel
 * \brief An assembler for the global Jacobian matrix for models using the box discretization.
 */
template<class TypeTag>
class BoxAssembler : public ImplicitAssembler<TypeTag>
{
    typedef ImplicitAssembler<TypeTag> ParentType;
    friend class ImplicitAssembler<TypeTag>;
    typedef typename GET_PROP_TYPE(TypeTag, GridView) GridView;
    typedef typename GET_PROP_TYPE(TypeTag, Scalar) Scalar;
    typedef typename GET_PROP_TYPE(TypeTag, VertexMapper) VertexMapper;
    typedef typename GET_PROP_TYPE(TypeTag, JacobianMatrix) JacobianMatrix;
    typedef typename GridView::template Codim<0>::Entity Element;
    typedef typename GridView::template Codim<0>::Iterator ElementIterator;
    enum{ dim = GridView::dimension };
    typedef typename GridView::template Codim<dim>::EntityPointer VertexPointer;

public:
    BoxAssembler(): ParentType() {}
    
private:
    // copying the jacobian assembler is not a good idea
    BoxAssembler(const BoxAssembler &);

    /*!
     * \brief Determine the colors of vertices and elements for partial
     *        reassembly given a relative tolerance.
     *
     * The following approach is used:
     *
     * - Set all vertices and elements to 'green'
     * - Mark all vertices as 'red' which exhibit an relative error above
     *   the tolerance
     * - Mark all elements which feature a 'red' vetex as 'red'
     * - Mark all vertices which are not 'red' and are part of a
     *   'red' element as 'yellow'
     * - Mark all elements which are not 'red' and contain a
     *   'yellow' vertex as 'yellow'
     *
     * \param relTol The relative error below which a vertex won't be
     *               reassembled. Note that this specifies the
     *               worst-case relative error between the last
     *               linearization point and the current solution and
     *               _not_ the delta vector of the Newton iteration!
     */
    void computeColors_(const Scalar relTol)
    {
        if (!this->enablePartialReassemble_())
            return;

        ElementIterator eIt = this->gridView_().template begin<0>();
        ElementIterator eEndIt = this->gridView_().template end<0>();

        // mark the red vertices and update the tolerance of the
        // linearization which actually will get achieved
        this->nextReassembleAccuracy_ = 0;
        for (unsigned int i = 0; i < this->vertexColor_.size(); ++i) {
            if (this->delta_[i] > relTol)
                // mark vertex as red if discrepancy is larger than
                // the relative tolerance
                this->vertexColor_[i] = ParentType::Red;
            else
                this->nextReassembleAccuracy_ =
                    std::max(this->nextReassembleAccuracy_, this->delta_[i]);
        }

        // Mark all red elements
        for (; eIt != eEndIt; ++eIt) {
            // find out whether the current element features a red
            // vertex
            bool isRed = false;

#if DUNE_VERSION_NEWER(DUNE_COMMON, 2, 4)
            int numVertices = eIt->subEntities(dim);
#else
            int numVertices = eIt->template count<dim>();
#endif
            for (int i=0; i < numVertices; ++i) {
#if DUNE_VERSION_NEWER(DUNE_COMMON, 2, 4)
                int globalI = this->vertexMapper_().subIndex(*eIt, i, dim);
#else
                int globalI = this->vertexMapper_().map(*eIt, i, dim);
#endif
                if (this->vertexColor_[globalI] == ParentType::Red) {
                    isRed = true;
                    break;
                }
            }

            // if yes, the element color is also red, else it is not
            // red, i.e. green for the mean time
#if DUNE_VERSION_NEWER(DUNE_COMMON, 2, 4)
            int eIdxGlobal = this->elementMapper_().index(*eIt);
#else
            int eIdxGlobal = this->elementMapper_().map(*eIt);
#endif
            if (isRed)
                this->elementColor_[eIdxGlobal] = ParentType::Red;
            else
                this->elementColor_[eIdxGlobal] = ParentType::Green;
        }

        // Mark yellow vertices (as orange for the mean time)
        eIt = this->gridView_().template begin<0>();
        for (; eIt != eEndIt; ++eIt) {
#if DUNE_VERSION_NEWER(DUNE_COMMON, 2, 4)
            int eIdx = this->elementMapper_().index(*eIt);
#else
            int eIdx = this->elementMapper_().map(*eIt);
#endif
            if (this->elementColor_[eIdx] != ParentType::Red)
                continue; // non-red elements do not tint vertices
                          // yellow!

#if DUNE_VERSION_NEWER(DUNE_COMMON, 2, 4)
            int numVertices = eIt->subEntities(dim);
#else
            int numVertices = eIt->template count<dim>();
#endif
            for (int i = 0; i < numVertices; ++i) {
#if DUNE_VERSION_NEWER(DUNE_COMMON, 2, 4)
                int globalI = this->vertexMapper_().subIndex(*eIt, i, dim);
#else
                int globalI = this->vertexMapper_().map(*eIt, i, dim);
#endif
                // if a vertex is already red, don't recolor it to
                // yellow!
                if (this->vertexColor_[globalI] != ParentType::Red) {
                    this->vertexColor_[globalI] = ParentType::Orange;
                }
            }
        }

        // at this point we communicate the yellow vertices to the
        // neighboring processes because a neigbor process may not see
        // the red vertex for yellow border vertices
        typedef typename ParentType::EntityColor EntityColor;
        VertexHandleMin<EntityColor, std::vector<EntityColor>,  VertexMapper>
            minHandle(this->vertexColor_, this->vertexMapper_());
        this->gridView_().communicate(minHandle,
                                Dune::InteriorBorder_InteriorBorder_Interface,
                                Dune::ForwardCommunication);

        // Mark yellow elements
        eIt = this->gridView_().template begin<0>();
        for (; eIt != eEndIt; ++eIt) {
#if DUNE_VERSION_NEWER(DUNE_COMMON, 2, 4)
            int eIdx = this->elementMapper_().index(*eIt);
#else
            int eIdx = this->elementMapper_().map(*eIt);
#endif
            if (this->elementColor_[eIdx] == ParentType::Red) {
                continue; // element is red already!
            }

            // check whether the element features a yellow
            // (resp. orange at this point) vertex
            bool isYellow = false;
#if DUNE_VERSION_NEWER(DUNE_COMMON, 2, 4)
            int numVertices = eIt->subEntities(dim);
#else
            int numVertices = eIt->template count<dim>();
#endif
            for (int i = 0; i < numVertices; ++i) {
#if DUNE_VERSION_NEWER(DUNE_COMMON, 2, 4)
                int globalI = this->vertexMapper_().subIndex(*eIt, i, dim);
#else
                int globalI = this->vertexMapper_().map(*eIt, i, dim);
#endif
                if (this->vertexColor_[globalI] == ParentType::Orange) {
                    isYellow = true;
                    break;
                }
            }

            if (isYellow)
                this->elementColor_[eIdx] = ParentType::Yellow;
        }

        // Demote orange vertices to yellow ones if it has at least
        // one green element as a neighbor.
        eIt = this->gridView_().template begin<0>();
        for (; eIt != eEndIt; ++eIt) {
#if DUNE_VERSION_NEWER(DUNE_COMMON, 2, 4)
            int eIdx = this->elementMapper_().index(*eIt);
#else
            int eIdx = this->elementMapper_().map(*eIt);
#endif
            if (this->elementColor_[eIdx] != ParentType::Green)
                continue; // yellow and red elements do not make
                          // orange vertices yellow!

#if DUNE_VERSION_NEWER(DUNE_COMMON, 2, 4)
            int numVertices = eIt->subEntities(dim);
#else
            int numVertices = eIt->template count<dim>();
#endif
            for (int i = 0; i < numVertices; ++i) {
#if DUNE_VERSION_NEWER(DUNE_COMMON, 2, 4)
                int globalI = this->vertexMapper_().subIndex(*eIt, i, dim);
#else
                int globalI = this->vertexMapper_().map(*eIt, i, dim);
#endif
                // if a vertex is orange, recolor it to yellow!
                if (this->vertexColor_[globalI] == ParentType::Orange)
                    this->vertexColor_[globalI] = ParentType::Yellow;
            }
        }

        // demote the border orange vertices
        VertexHandleMax<EntityColor, std::vector<EntityColor>,  VertexMapper>
            maxHandle(this->vertexColor_,
                      this->vertexMapper_());
        this->gridView_().communicate(maxHandle,
                                Dune::InteriorBorder_InteriorBorder_Interface,
                                Dune::ForwardCommunication);

        // promote the remaining orange vertices to red
        for (unsigned int i=0; i < this->vertexColor_.size(); ++i) {
            // if a vertex is green or yellow don't do anything!
            if (this->vertexColor_[i] == ParentType::Green || this->vertexColor_[i] == ParentType::Yellow)
                continue;

            // make sure the vertex is red (this is a no-op vertices
            // which are already red!)
            this->vertexColor_[i] = ParentType::Red;

            // set the error of this vertex to 0 because the system
            // will be consistently linearized at this vertex
            this->delta_[i] = 0.0;
        }
    }

    // Construct the BCRS matrix for the global jacobian
    void createMatrix_()
    {
        int numVerticesGlobal = this->gridView_().size(dim);

        // allocate raw matrix
        this->matrix_ = Dune::make_shared<JacobianMatrix>(numVerticesGlobal, numVerticesGlobal, JacobianMatrix::random);

        // find out the global indices of the neighboring vertices of
        // each vertex
        typedef std::set<int> NeighborSet;
        std::vector<NeighborSet> neighbors(numVerticesGlobal);
        ElementIterator eIt = this->gridView_().template begin<0>();
        const ElementIterator eEndIt = this->gridView_().template end<0>();
        for (; eIt != eEndIt; ++eIt) {
            const Element &element = *eIt;
#if DUNE_VERSION_NEWER(DUNE_COMMON, 2, 4)
            int numVerticesLocal = element.subEntities(dim);
#else
            int numVerticesLocal = element.template count<dim>();
#endif

            // if the element is not in the interior or the process
            // border, all dofs just contain main-diagonal entries
            if (element.partitionType() != Dune::InteriorEntity &&
                element.partitionType() != Dune::BorderEntity)
            {
                for (int i = 0; i < numVerticesLocal; ++i) {
#if DUNE_VERSION_NEWER(DUNE_COMMON, 2, 4)
                    int globalI = this->vertexMapper_().subIndex(*eIt, i, dim);
#else
                    int globalI = this->vertexMapper_().map(*eIt, i, dim);
#endif
                    neighbors[globalI].insert(globalI);
                }
            }
            else
            {
                // loop over all element vertices
                for (int i = 0; i < numVerticesLocal - 1; ++i) {
#if DUNE_VERSION_NEWER(DUNE_COMMON, 2, 4)
                    int globalI = this->vertexMapper_().subIndex(*eIt, i, dim);
#else
                    int globalI = this->vertexMapper_().map(*eIt, i, dim);
#endif
                    for (int j = i + 1; j < numVerticesLocal; ++j) {
#if DUNE_VERSION_NEWER(DUNE_COMMON, 2, 4)
                        int globalJ = this->vertexMapper_().subIndex(*eIt, j, dim);
#else
                        int globalJ = this->vertexMapper_().map(*eIt, j, dim);
#endif
                        // make sure that vertex j is in the neighbor set
                        // of vertex i and vice-versa
                        neighbors[globalI].insert(globalJ);
                        neighbors[globalJ].insert(globalI);
                    }
                }
            }
        }

        // make vertices neighbors to themselfs
        for (int i = 0; i < numVerticesGlobal; ++i) {
            neighbors[i].insert(i);
        }
        
        // allocate space for the rows of the matrix
        for (int i = 0; i < numVerticesGlobal; ++i) {
            this->matrix_->setrowsize(i, neighbors[i].size());
        }
        this->matrix_->endrowsizes();

        // fill the rows with indices. each vertex talks to all of its
        // neighbors. (it also talks to itself since vertices are
        // sometimes quite egocentric.)
        for (int i = 0; i < numVerticesGlobal; ++i) {
            typename NeighborSet::iterator nIt = neighbors[i].begin();
            typename NeighborSet::iterator nEndIt = neighbors[i].end();
            for (; nIt != nEndIt; ++nIt) {
                this->matrix_->addindex(i, *nIt);
            }
        }
        this->matrix_->endindices();
    }

    // assemble a non-ghost element
    void assembleElement_(const Element &element)
    {
        if (this->enablePartialReassemble_()) {
#if DUNE_VERSION_NEWER(DUNE_COMMON, 2, 4)
            int eIdxGlobal = this->model_().elementMapper().index(element);
#else
            int eIdxGlobal = this->model_().elementMapper().map(element);
#endif
            if (this->elementColor_[eIdxGlobal] == ParentType::Green) {
                ++this->greenElems_;

                assembleGreenElement_(element);
                return;
            }
        }

        this->model_().localJacobian().assemble(element);

#if DUNE_VERSION_NEWER(DUNE_COMMON, 2, 4)
        int numVerticesLocal = element.subEntities(dim);
#else
        int numVerticesLocal = element.template count<dim>();
#endif
        for (int i=0; i < numVerticesLocal; ++ i) {
#if DUNE_VERSION_NEWER(DUNE_COMMON, 2, 4)
            int globI = this->vertexMapper_().subIndex(element, i, dim);
#else
            int globI = this->vertexMapper_().map(element, i, dim);
#endif

            // update the right hand side
            this->residual_[globI] += this->model_().localJacobian().residual(i);
            for (int j = 0; j < this->residual_[globI].dimension; ++j) {
                if (!std::isfinite(this->residual_[globI][j])) {
                    DUNE_THROW(NumericalProblem, 
                               "residual_[" << globI << "][" << j << "] is not finite");
                }
            }
            if (this->enableJacobianRecycling_()) {
                this->storageTerm_[globI] +=
                    this->model_().localJacobian().storageTerm(i);
            }

            // only update the jacobian matrix for non-green vertices
            if (this->vertexColor(globI) != ParentType::Green) {
                if (this->enableJacobianRecycling_())
                    this->storageJacobian_[globI] +=
                        this->model_().localJacobian().storageJacobian(i);

                // update the jacobian matrix
                for (int j = 0; j < numVerticesLocal; ++ j) {
#if DUNE_VERSION_NEWER(DUNE_COMMON, 2, 4)
                    int globJ = this->vertexMapper_().subIndex(element, j, dim);
#else
                    int globJ = this->vertexMapper_().map(element, j, dim);
#endif
                    (*this->matrix_)[globI][globJ] +=
                        this->model_().localJacobian().mat(i,j);
                }
            }
        }
    }

    // "assemble" a green element. green elements only get the
    // residual updated, but the jacobian is left alone...
    void assembleGreenElement_(const Element &element)
    {
        this->model_().localResidual().eval(element);

#if DUNE_VERSION_NEWER(DUNE_COMMON, 2, 4)
        int numVerticesLocal = element.subEntities(dim);
#else
        int numVerticesLocal = element.template count<dim>();
#endif
        for (int i = 0; i < numVerticesLocal; ++ i) {
#if DUNE_VERSION_NEWER(DUNE_COMMON, 2, 4)
            int globI = this->vertexMapper_().subIndex(element, i, dim);
#else
            int globI = this->vertexMapper_().map(element, i, dim);
#endif

            // update the right hand side
            this->residual_[globI] += this->model_().localResidual().residual(i);
            if (this->enableJacobianRecycling_())
                this->storageTerm_[globI] += this->model_().localResidual().storageTerm(i);
        }
    }

    // "assemble" a ghost element
    void assembleGhostElement_(const Element &element)
    {
#if DUNE_VERSION_NEWER(DUNE_COMMON, 2, 4)
        int numVerticesLocal = element.subEntities(dim);
#else
        int numVerticesLocal = element.template count<dim>();
#endif
        for (int i=0; i < numVerticesLocal; ++i) {
            const VertexPointer vp = element.template subEntity<dim>(i);

            if (vp->partitionType() == Dune::InteriorEntity ||
                vp->partitionType() == Dune::BorderEntity)
            {
                // do not change the non-ghost vertices
                continue;
            }

            // set main diagonal entries for the vertex
#if DUNE_VERSION_NEWER(DUNE_COMMON, 2, 4)
            int vIdx = this->vertexMapper_().index(*vp);
#else
            int vIdx = this->vertexMapper_().map(*vp);
#endif
            typedef typename JacobianMatrix::block_type BlockType;
            BlockType &J = (*this->matrix_)[vIdx][vIdx];
            for (int j = 0; j < BlockType::rows; ++j)
                J[j][j] = 1.0;

            // set residual for the vertex
            this->residual_[vIdx] = 0;
        }
    }
};

} // namespace Dumux

#endif
