// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
// vi: set et ts=4 sw=4 sts=4:
/*****************************************************************************
 *   See the file COPYING for full copying permissions.                      *
 *                                                                           *
 *   This program is free software: you can redistribute it and/or modify    *
 *   it under the terms of the GNU General Public License as published by    *
 *   the Free Software Foundation, either version 2 of the License, or       *
 *   (at your option) any later version.                                     *
 *                                                                           *
 *   This program is distributed in the hope that it will be useful,         *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the            *
 *   GNU General Public License for more details.                            *
 *                                                                           *
 *   You should have received a copy of the GNU General Public License       *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.   *
 *****************************************************************************/
/*!
 * \file
 * \brief Writes the VTK output files for a solution of the MpNc model.
 */

#ifndef DUMUX_MPNC_VTK_WRITER_HH
#define DUMUX_MPNC_VTK_WRITER_HH

#include "mpncproperties.hh"

#include <dumux/io/vtkmultiwriter.hh>

namespace Dumux
{
/*!
 * \ingroup MPNCModel
 * \brief Writes the VTK output files for a
 * solution of the MpNc model.
 */
template<class TypeTag>
class MPNCVtkWriter
{
    typedef typename GET_PROP_TYPE(TypeTag, MPNCVtkCommonModule) MPNCVtkCommonModule;
    typedef typename GET_PROP_TYPE(TypeTag, MPNCVtkMassModule) MPNCVtkMassModule;
    typedef typename GET_PROP_TYPE(TypeTag, MPNCVtkEnergyModule) MPNCVtkEnergyModule;
    typedef typename GET_PROP_TYPE(TypeTag, MPNCVtkCustomModule) MPNCVtkCustomModule;
    typedef typename GET_PROP_TYPE(TypeTag, ElementVolumeVariables) ElementVolumeVariables;
    typedef typename GET_PROP_TYPE(TypeTag, FVElementGeometry) FVElementGeometry;
    typedef typename GET_PROP_TYPE(TypeTag, ElementBoundaryTypes) ElementBoundaryTypes;
    typedef typename GET_PROP_TYPE(TypeTag, Problem) Problem;
    typedef typename GET_PROP_TYPE(TypeTag, GridView) GridView;
    typedef typename GridView::template Codim<0>::Iterator ElementIterator;

public:
    MPNCVtkWriter(const Problem &problem)
        : problem_(problem)
        , commonWriter_(problem)
        , massWriter_(problem)
        , energyWriter_(problem)
        , customWriter_(problem)
    {
    }

    /*!
     * \brief Add  the current solution to the VtkMultiWriter.
     */
    template <class MultiWriter>
    void addCurrentSolution(MultiWriter &writer)
    {
        // tell sub-writers to allocate their buffers
        commonWriter_.allocBuffers(writer);
        massWriter_.allocBuffers(writer);
        energyWriter_.allocBuffers(writer);
        customWriter_.allocBuffers(writer);

        // iterate over grid
        FVElementGeometry fvGeometry;
        ElementVolumeVariables elemVolVars;
        ElementBoundaryTypes elemBcTypes;

        ElementIterator eIt = problem_.gridView().template begin<0>();
        ElementIterator eEndIt = problem_.gridView().template end<0>();
        for (; eIt != eEndIt; ++eIt)
        {
            if(eIt->partitionType() == Dune::InteriorEntity)
            {
                fvGeometry.update(problem_.gridView(), *eIt);
                elemBcTypes.update(problem_, *eIt);
                this->problem_.model().setHints(*eIt, elemVolVars);
                elemVolVars.update(problem_,
                                   *eIt,
                                   fvGeometry,
                                   false);
                this->problem_.model().updateCurHints(*eIt, elemVolVars);

                // tell the sub-writers to do what ever they need to with
                // their internal buffers when a given element is seen.
                commonWriter_.processElement(*eIt,
                                             fvGeometry,
                                             elemVolVars,
                                             elemBcTypes);
                massWriter_.processElement(*eIt,
                                           fvGeometry,
                                           elemVolVars,
                                           elemBcTypes);
                energyWriter_.processElement(*eIt,
                                             fvGeometry,
                                             elemVolVars,
                                             elemBcTypes);
                customWriter_.processElement(*eIt,
                                             fvGeometry,
                                             elemVolVars,
                                             elemBcTypes);
            }
        }

        // write everything to the output file
        commonWriter_.commitBuffers(writer);
        massWriter_.commitBuffers(writer);
        energyWriter_.commitBuffers(writer);
        customWriter_.commitBuffers(writer);
    }

private:
    const Problem &problem_;

    MPNCVtkCommonModule commonWriter_;
    MPNCVtkMassModule massWriter_;
    MPNCVtkEnergyModule energyWriter_;
    MPNCVtkCustomModule customWriter_;
};

}

#endif
