// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
// vi: set et ts=4 sw=4 sts=4:
/*****************************************************************************
 *   See the file COPYING for full copying permissions.                      *
 *                                                                           *
 *   This program is free software: you can redistribute it and/or modify    *
 *   it under the terms of the GNU General Public License as published by    *
 *   the Free Software Foundation, either version 2 of the License, or       *
 *   (at your option) any later version.                                     *
 *                                                                           *
 *   This program is distributed in the hope that it will be useful,         *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the            *
 *   GNU General Public License for more details.                            *
 *                                                                           *
 *   You should have received a copy of the GNU General Public License       *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.   *
 *****************************************************************************/
/*!
 * \file
 *
 * \brief   Parameters that are necessary for the \em regularization of
 *          VanGenuchten "material law".
 */

#ifndef REGULARIZED_VAN_GENUCHTEN_PARAMS_HH
#define REGULARIZED_VAN_GENUCHTEN_PARAMS_HH

#include "vangenuchtenparams.hh"

namespace Dumux
{
/*!
 *
 *
 * \brief   Parameters that are necessary for the \em regularization of
 *          VanGenuchten "material law".
 *
 * \ingroup fluidmatrixinteractionsparams
 */
template<class ScalarT>
class RegularizedVanGenuchtenParams : public VanGenuchtenParams<ScalarT>
{
public:
    typedef ScalarT Scalar;
    typedef VanGenuchtenParams<Scalar> Parent;

    RegularizedVanGenuchtenParams()
    {}

    RegularizedVanGenuchtenParams(Scalar vgAlpha,
                                  Scalar vgN)
        : Parent(vgAlpha, vgN)
    {}

    /*!
     * \brief Threshold saturation below which the capillary pressure
     *        is regularized.
     *
     * This is just 1%. If you need a different value, overload this
     * class.
     */
    Scalar pcLowSw() const
    {
        // Most problems are very sensitive to this value
        // (e.g. making it smaller might result in negative
        // pressures)
        //
        // If you want to use a different regularization threshold,
        // overload this class and supply the new class as second
        // template parameter for the RegularizedVanGenuchten law!
        return 1e-2;
    }

    /*!
     * \brief Threshold saturation above which the capillary pressure
     *        is regularized.
     *
     * This is just 99%. If you need a different value, overload this
     * class.
     */
    Scalar pcHighSw() const
    {
        // Most problems are very sensitive to this value
        // (e.g. making it smaller might result in negative
        // pressures)
        //
        // If you want to use a different regularization threshold,
        // overload this class and supply the new class as second
        // template parameter for the RegularizedVanGenuchten law!
        return 99e-2;
    }

    /*!
     * \brief Threshold saturation below which the relative
     *        permeability of the non-wetting phase gets regularized.
     *
     * This is just 10%. If you need a different value, overload this
     * class.
     */
    Scalar krnLowSw() const
    { return 0.10; }

    /*!
     * \brief Threshold saturation above which the relative
     *        permeability of the wetting phase gets regularized.
     *
     * This is just 90%. If you need a different value, overload this
     * class.
     */
    Scalar krwHighSw() const
    { return 0.90; }

};
} // namespace Dumux

#endif
